// lemming-robo-firmware-v0.1





#include "common/config.h"
#include "defs.h"
#include "common/i2c.h"
#include "common/imu.h"
#include "common/motor.h"
#include "common/pins.h"
#include "common/sonar.h"
#include "common/tools.h"

#include <stdio.h>

// These are all open source libraries written by Kristian Lauszus, TKJ Electronics
// The USB libraries are located at the following link: https://github.com/felis/USB_Host_Shield_2.0
#include <Kalman.h> // Kalman filter library - see: http://blog.tkjelectronics.dk/2012/09/a-practical-approach-to-kalman-filter-and-how-to-implement-it/

// Sonar objects:
Sonar_Object _sonarL; Sonar_Object* sonarL = &_sonarL;
Sonar_Object _sonarR; Sonar_Object* sonarR = &_sonarR;

// Create the Kalman library instance
Kalman kalman; // See https://github.com/TKJElectronics/KalmanFilter for source code

float omegaVersatz_querachse; // in LSB
float omegaVersatz_hochachse; // in LSB, Hochachse dient der ausregelung des Geradeausfahrens. 


/* Used for timing */
static uint32_t kalmanTimer; // Timer used for the Kalman filter
static uint32_t pidTimer; // Timer used for the PID loop
static uint32_t imuTimer; // This is used to set a delay between sending IMU values
static uint32_t encoderTimer; // Timer used to determine when to update the encoder values
static uint32_t reportTimer; // This is used to set a delay between sending report values
static uint32_t ledTimer; // Used to update the LEDs to indicate battery level on the PS3, PS4, Wii and Xbox controllers
static uint32_t blinkTimer; // Used to blink the built in LED, starts blinking faster upon an incoming Bluetooth request

static bool layingDown = true; // Use to indicate if the robot is laying down

bool calibrateGyro();
float winkelDesBeschleunigungsvektors();
float drehgeschwindigkeitUmQuerachse();
float drehgeschwindigkeitUmHochachse();
bool drehgeschwindigkeitUmQuerachseKalibrieren();
bool drehgeschwindigkeitUmHochachseKalibrieren();
bool checkMinMax(int16_t *array, uint8_t length, int16_t maxDifference);
float parameterEinstellen();
float lenken();

float winkelRegler(float dt);
float positionsRegler(float dt);

// Einmalig alles einrichten
void setup() // dieser Funktionsname darf nicht verändert werden!
{
	Serial.begin(115200); // Serielle Schnittstelle über USB zum Computer einrichten

	CFG_init();

	IMU_einrichten();
	Pins_einrichten();
	Sonar_init(sonarL, Sonar_L_Trig_Pin, Sonar_L_Echo_Pin);
	Sonar_init(sonarR, Sonar_R_Trig_Pin, Sonar_R_Echo_Pin);

	/* Set Kalman and gyro starting angle */
	IMU_werteEinlesen();
	P.phiBeschl.v = winkelDesBeschleunigungsvektors();

	kalman.setAngle(P.phiBeschl.v); // Set starting angle
	P.phiKalman.v = P.phiBeschl.v;
	P.phiGyro.v = P.phiBeschl.v;

	/* Calibrate gyro zero value */

	Motor_beep(1000, 100);
	delay(500);
	Motor_beep(1000, 100);
	delay(500);
	Motor_beep(1000, 100);
	delay(500);
	Motor_beep(1000, 500);



	while (drehgeschwindigkeitUmQuerachseKalibrieren()); // Run again if the robot is moved while calibrating
	while(drehgeschwindigkeitUmHochachseKalibrieren());

	Motor_beep(1000, 100);
	delay(100);
	Motor_beep(2000, 300);



	#ifdef ENABLE_TOOLS
		printMenu();
	#endif


	/* Setup timing */
	kalmanTimer = micros();
	pidTimer = kalmanTimer;
	imuTimer = millis();
	encoderTimer = imuTimer;
	reportTimer = imuTimer;
	ledTimer = imuTimer;
	blinkTimer = imuTimer;
}

// Endlosschleife
void loop() // dieser Funktionsname darf nicht verändert werden!
{
	/* Calculate winkel */
	IMU_werteEinlesen();

	int16_t omega = drehgeschwindigkeitUmQuerachse(); // in LSB
	int16_t aHoch = imu.aY; 
	int16_t aLaengs = -imu.aZ;

	// atan2 outputs the value of -π to π (radians) - see http://en.wikipedia.org/wiki/Atan2
	// We then convert it to 0 to 2π and then from radians to degrees
	P.phiBeschl.v = winkelDesBeschleunigungsvektors();

	uint32_t timer = micros();
	// This fixes the 0-360 transition problem when the accelerometer angle jumps between 0 and 360 degrees
	if ((P.phiBeschl.v < 90 && P.phiKalman.v > 270) || (P.phiBeschl.v > 270 && P.phiKalman.v < 90)) {
		kalman.setAngle(P.phiBeschl.v);
		P.phiKalman.v = P.phiBeschl.v;
		P.phiGyro.v = P.phiBeschl.v;
		P.phiSelekt.v = P.phiBeschl.v;
	} else {
		P.omegaGyro.v = ((float)omega - omegaVersatz_querachse) / 131.0f; // Convert to deg/s
		float dt = (float)(timer - kalmanTimer) / 1000000.0f;
		P.phiGyro.v += P.omegaGyro.v * dt; // Gyro angle is only used for debugging
		if (P.phiGyro.v < 0 || P.phiGyro.v > 360)
			P.phiGyro.v = P.phiKalman.v; // Reset the gyro angle when it has drifted too much
		P.phiKalman.v = kalman.getAngle(P.phiBeschl.v, P.omegaGyro.v, dt); // Calculate the angle using a Kalman filter
		P.omegaKalman.v = kalman.getRate();

		selektiverFilter(dt); // resultat in P.phiSelekt

	}
	kalmanTimer = timer;
	//Serial.print(P.phiBeschl.v);Serial.print('\t');Serial.print(P.phiGyro.v);Serial.print('\t');Serial.println(winkel);

	/* Drive motors */

	// If the robot is laying down, it has to be put in a vertical position (+-10°) before it starts balancing
	// If it's already balancing it has to be ±45 degrees before it stops trying to balance
	if ((layingDown && (P.phiKalman.v < P.phiNull.v - 1 || P.phiKalman.v > P.phiNull.v + 1)) || 
		(!layingDown && (P.phiKalman.v < P.phiNull.v - 45 || P.phiKalman.v > P.phiNull.v + 45))) {
		layingDown = true; // The robot is in a unsolvable position, so turn off both motors and wait until it's vertical again
		Motor_L(0);
		Motor_R(0);
	} else {


		timer = micros();
		float dt = (float)(timer-pidTimer)*1e-6;

		// Verzögerung, sodass minimale Abtastzeit eingehalten wird
		// (zu schnelles Reglen führt zu numerischen instabilitäten)
		while(dt < P.dtMin.v)
		{
			timer = micros();
			dt = (float)(timer-pidTimer)*1e-6;
		}

		layingDown = false; // It's no longer laying down

		P.pwmMotor.v = winkelRegler(dt); // unterlagerter Regler
		P.phiSteuer.v = positionsRegler4(dt); // überlagerter Regler

		pidTimer = timer;

		float omegaHoch = (drehgeschwindigkeitUmHochachse()-omegaVersatz_hochachse)/131.0; // in °/s, positiv = linkskurve, negativ = rechtskurve

		float r = 95.0e-3/2; // Rad Achsabstand in m
		float omegaMax = P.kMotor.v*r / (2*M_PI)*360; // max. hochachs rotationsgeschwindigkeit in °/s
			// wenn ein motor voll rückwärts und einer voll vorwärts dreht. 

		float omegaSoll = lenken(); // °/s; (vorerst nur geradeaus fahren)
		float omegaFehler = omegaSoll - omegaHoch; 
		static float omegaFehlerIntegral = 0;
		omegaFehlerIntegral += omegaFehler*dt; // in °

		float pwmL = P.pwmMotor.v*255.0f - (omegaFehler*0.2 + omegaFehlerIntegral);
		float pwmR = P.pwmMotor.v*255.0f + (omegaFehler*0.2 + omegaFehlerIntegral);

		Motor_L(pwmL * P.motorEin.v);
		Motor_R(pwmR * P.motorEin.v);

		CFG_printTableRow(dt);


		// 10s/U offset. soll innerhalb von 5s ausgeglichen sein. 
		// omega = 36°/s
		// richtungsVersatz soll dann ca. 15 sein. 
		// 15 = 5s * 36°/s * k *dt
		//richtungsVersatz += 0.1*drehgeschwindigkeitUmHochachse()/(1<<15)/200*dt; // TODO: Hochachsenregler!
	}
	parameterEinstellen();
}













// dt: Zeitschritt in s seit letztem Aufruf
// gibt phiSteuer in ° zurück
float positionsRegler(float dt)
{
	static float pwmGefiltert=0;

	if(dt > 0.5)
	{
		P.pos.v = 0; // Integrator zurücksetzten, wenn Robo gelegen ist. 
		pwmGefiltert = 0; // Filter zurücksetzten, wenn Robo gelegen ist. 
	}
	else // nach einer pause nicht integrieren!
	{
		pwmGefiltert = pwmGefiltert + (P.pwmMotor.v - pwmGefiltert)*P.pwmFilter_tau.v*dt; // IIR-Filter 1.Ordnung
		P.vMotor.v = pwmGefiltert * P.kMotor.v; // Umrechnung auf m/s
		P.pos.v += P.vMotor.v*dt; // integrierte Geschwindigkeit ergibt Position
	}

	P.posFehler.v = P.posSoll.v - P.pos.v;

	float winkel=0;
	static float posFehlerAlt = 0; // wird für das Differenzieren benötigt. 

	winkel += P.posReg_P.v * P.posFehler.v; // P-Anteil
	winkel = constrain(winkel, -P.phiLimit.v/4, +P.phiLimit.v/4); 
		// P-Anteil (Feder) darf bei großen Abweichungen nicht dominieren!

	float dX_dt = (P.posFehler.v - posFehlerAlt)/dt;

	winkel += P.posReg_D.v * dX_dt; // D-Anteil
	posFehlerAlt = P.posFehler.v;

	P.phiSteuer.v = constrain(winkel, -P.phiLimit.v, +P.phiLimit.v);
	// if(P.phiSteuer.v != winkel)
	// 	P.pos.v -= P.vMotor.v*dt; // Integration rückgängig machen, wenn limit erreicht ist. 

	return P.phiSteuer.v;
}

// Positionsregler wie bei Balanduino mit verschiedenen Zonen und P-Faktoren. 
float positionsRegler2(float dt)
{
	static float pwmGefiltert=0;

	if(dt > 0.5)
	{
		P.pos.v = 0; // Integrator zurücksetzten, wenn Robo gelegen ist. 
		pwmGefiltert = 0; // Filter zurücksetzten, wenn Robo gelegen ist. 
	}
	else // nach einer pause nicht integrieren!
	{
		pwmGefiltert = pwmGefiltert + (P.pwmMotor.v - pwmGefiltert)*P.pwmFilter_tau.v*dt; // IIR-Filter 1.Ordnung
		P.vMotor.v = pwmGefiltert * P.kMotor.v; // Umrechnung auf m/s
		P.pos.v += P.vMotor.v*dt; // integrierte Geschwindigkeit ergibt Position
	}

	P.posFehler.v = P.posSoll.v - P.pos.v;

	float winkel=0;
	static float posFehlerAlt = 0; // wird für das Differenzieren benötigt. 

	// TODO: manipulierbar machen. 
	float zoneC = 0.15; // in m, engster Bereich
	float zoneB = 4*zoneC; // m
	float zoneA = 8*zoneC; // m, weitester Bereich


	if (fabs(P.posFehler.v) > zoneA) // Inside zone A
		winkel += P.posReg_P.v * P.posFehler.v * 0.33; // P-Anteil
	else if (fabs(P.posFehler.v) > zoneB) // Inside zone B
		winkel += P.posReg_P.v * P.posFehler.v * 0.25; // P-Anteil
	else if (fabs(P.posFehler.v) > zoneC) // Inside zone C
		winkel += P.posReg_P.v * P.posFehler.v * 0.20; // P-Anteil
	else // Inside zone D
		winkel += P.posReg_P.v * P.posFehler.v * 1.0; // P-Anteil

	float dX_dt = (P.posFehler.v - posFehlerAlt)/dt;

	winkel += P.posReg_D.v * dX_dt; // D-Anteil
	posFehlerAlt = P.posFehler.v;

	static float winkelAlt=0;
	winkel = constrain(winkel, -P.phiLimit.v, +P.phiLimit.v);
	winkel = constrain(winkel, winkelAlt-1, winkelAlt+1); // Beschränkung der Veränderungsrate des Winkels auf 1°/Aufruf

	P.phiSteuer.v = winkel;
	winkelAlt = winkel;

	return P.phiSteuer.v;
}

// dt: Zeitschritt in s seit letztem Aufruf
// gibt phiSteuer in ° zurück
// positionsRegler() führt zum Aufschwingen der Position!
// mit diesem Regler wird experimentiert. 
float positionsRegler3(float dt)
{
	static float distGefiltert=0;

	if(dt > 0.5)
	{
		P.pos.v = 0; // Integrator zurücksetzten, wenn Robo gelegen ist. 
		distGefiltert = 0.4; // Filter zurücksetzten, wenn Robo gelegen ist. 
	}
	else // nach einer pause nicht integrieren!
	{
		Sonar_measure(sonarL, 700);
		Sonar_measure(sonarR, 700);
		float dist;

		if(sonarL->distance_mm >200 && sonarL->distance_mm < 600 &&
		   sonarR->distance_mm >200 && sonarR->distance_mm < 600)
		{
			dist = ((float)(sonarL->distance_mm + sonarR->distance_mm))/2e3; // distance in m	
		} 
		else if(sonarL->distance_mm >200 && sonarL->distance_mm < 600)
		{
			dist = ((float)(sonarL->distance_mm))/1.0e3; // distance in m	
		} 
		else if(sonarR->distance_mm >200 && sonarR->distance_mm < 600)
		{
			dist = ((float)(sonarR->distance_mm))/1.0e3; // distance in m
		}else{
			dist = 0.4;
		}
	
		distGefiltert = distGefiltert + (dist - distGefiltert)*P.pwmFilter_tau.v*dt; 
			// IIR-Filter 1.Ordnung, hier wird der PWM-Filter für die Abstandsmessung hergenommen
		P.pos.v = -dist + 0.4; // Null-Punkt der Poition ist bei 40cm Abstand
	}

	P.posFehler.v = P.posSoll.v - P.pos.v;

	float winkel=0;
	static float posFehlerAlt = 0; // wird für das Differenzieren benötigt. 
	float e = P.posFehler.v;

	float dX_dt = (P.posFehler.v - posFehlerAlt)/dt;

// zuerst D-Anteil, dann begrenzen, dann P-Anteil:
	winkel += P.posReg_D.v * dX_dt; // D-Anteil
	winkel = constrain(winkel, -P.phiLimit.v, +P.phiLimit.v); 
		// P-Anteil (Feder) darf bei großen Abweichungen nicht dominieren!
	winkel += P.posReg_P.v * e; // P-Anteil

	posFehlerAlt = e;

	P.phiSteuer.v = constrain(winkel, -P.phiLimit.v, +P.phiLimit.v);
	// if(P.phiSteuer.v != winkel)
	// 	P.pos.v -= P.vMotor.v*dt; // Integration rückgängig machen, wenn limit erreicht ist. 

	return P.phiSteuer.v;
}

// dt: Zeitschritt in s seit letztem Aufruf
// gibt phiSteuer in ° zurück
float positionsRegler4(float dt)
{
	static float pwmGefiltert=0;
	//float vSoll = 0.1; // m/s

	if(dt > 0.5)
	{
		P.pos.v = 0; // Integrator zurücksetzten, wenn Robo gelegen ist. 
		pwmGefiltert = 0; // Filter zurücksetzten, wenn Robo gelegen ist. 
		P.posSoll.v = posSoll(0, true);
		return 0;
	}
	else // nach einer pause nicht integrieren!
	{
		pwmGefiltert = pwmGefiltert + (P.pwmMotor.v - pwmGefiltert)*P.pwmFilter_tau.v*dt; // IIR-Filter 1.Ordnung
		P.vMotor.v = pwmGefiltert * P.kMotor.v; // Umrechnung auf m/s
		P.pos.v += P.vMotor.v*dt; // integrierte Geschwindigkeit ergibt Position
	}

	P.posSoll.v = posSoll(dt, false);

	P.posFehler.v = P.posSoll.v - P.pos.v;

	float winkel=0;
	static float posFehlerAlt = 0; // wird für das Differenzieren benötigt. 

	winkel += P.posReg_P.v * P.posFehler.v; // P-Anteil
	winkel = constrain(winkel, -P.phiLimit.v/4, +P.phiLimit.v/4); 
		// P-Anteil (Feder) darf bei großen Abweichungen nicht dominieren!

	float dX_dt = (P.posFehler.v - posFehlerAlt)/dt;
	static float dX_dt_alt = 0;
	float dX2_dt2 = (dX_dt - dX_dt_alt)/dt;

	winkel += P.posReg_D.v * dX_dt; // D-Anteil
	winkel += P.posReg_DD.v * dX_dt; // D²-Anteil

	posFehlerAlt = P.posFehler.v;
	dX_dt_alt = dX_dt;

	P.phiSteuer.v = constrain(winkel, -P.phiLimit.v, +P.phiLimit.v);
	// if(P.phiSteuer.v != winkel)
	// 	P.pos.v -= P.vMotor.v*dt; // Integration rückgängig machen, wenn limit erreicht ist. 

	return P.phiSteuer.v;
}

// winkel: aktueller Lagewinkel (vom Kalmanfilter) in °
// dt: Zeitschritt in s seit letztem Aufruf
// gibt PWM-Wert zurück (-255...+255)
float winkelRegler(float dt)
{
	static float winkelFehlerAlt=0; 
	static float winkelFehlerIntegral=0;


	// Integrator zurücksetzten, wenn Regler zu lange inaktiv war!
	// (Robo ist gelegen und gerade aufgestanden)
	if(dt > 0.5)
		winkelFehlerIntegral = 0; 

	//P.phiFehler.v = -(P.phiNull.v + P.phiSteuer.v) + P.phiKalman.v; // in °
	//P.phiFehler.v = -(P.phiNull.v + P.phiSteuer.v) + P.phiGyro.v; // in °
	P.phiFehler.v = -(P.phiNull.v + P.phiSteuer.v) + P.phiSelekt.v; // in °
	winkelFehlerIntegral += P.phiFehler.v;


	float pwm = 0;
	pwm += P.phiReg_P.v * P.phiFehler.v; // P-Anteil
	//pwm += P.phiReg_I.v * winkelFehlerIntegral; // I-Anteil // TODO: I-Anteil mit limit einbauen. 
	pwm += P.phiReg_D.v * (P.phiFehler.v - winkelFehlerAlt); // D-Anteil

	P.pwmMotor.v = constrain(pwm, -1.0, +1.0);
	// TODO: antiwindup

	winkelFehlerAlt = P.phiFehler.v;

	return P.pwmMotor.v;
}




















// Definition der Koordinaten (abhängig von der Montage des MPU6050 auf dem Robo):
// Hochachse: y, positiv nach oben
// Querachse: x, positiv nach rechts
// Längsachse: z, positiv nach hinten

// Obige Achsen werden übersetzt auf folgende Achsen (fix)
// Laengs (positiv nach vorne)
// Hoch (positiv nach oben)

// Winkel: 
// bei aufrechter Position: 180°
// auf der Nase liegend: 270°
// auf dem Rücken liegend: 90°

float winkelDesBeschleunigungsvektors()
{
		/* Set Kalman and gyro starting angle */
	IMU_werteEinlesen();
	int16_t aHoch = imu.aX; 
	int16_t aLaengs = imu.aZ;
	// atan2 outputs the value of -π to π (radians) - see http://en.wikipedia.org/wiki/Atan2
	// We then convert it to 0 to 2π and then from radians to degrees
	return  
		(
			atan2
			(
				(float)aLaengs /*- P.aLaengsachseNullVersatz.v*/, 
				-((float)aHoch /*- P.aHochachseNullVersatz.v)*/) // TOOD
			) 
			+ PI
		) * RAD_TO_DEG;

}

// TODO: eindeutschen
bool drehgeschwindigkeitUmQuerachseKalibrieren() {
	int16_t gyroBuffer[25];
	for (uint8_t i = 0; i < 25; i++) {
		IMU_werteEinlesen();
		gyroBuffer[i] = drehgeschwindigkeitUmQuerachse();
		delay(10);
	}
	if (!checkMinMax(gyroBuffer, 25, 2000)) {
	  Serial.println(F("Gyro calibration error"));
	  return 1;
	} 
	for (uint8_t i = 0; i < 25; i++)
		omegaVersatz_querachse += gyroBuffer[i];
		omegaVersatz_querachse /= 25.0f;
	return 0;
}

bool drehgeschwindigkeitUmHochachseKalibrieren() {
	int16_t gyroBuffer[25];
	for (uint8_t i = 0; i < 25; i++) {
		IMU_werteEinlesen();
		gyroBuffer[i] = drehgeschwindigkeitUmHochachse();
		delay(10);
	}
	if (!checkMinMax(gyroBuffer, 25, 2000)) {
	  Serial.println(F("Gyro calibration error"));
	  return 1;
	} 
	for (uint8_t i = 0; i < 25; i++)
		omegaVersatz_hochachse += gyroBuffer[i];
		omegaVersatz_hochachse /= 25.0f;
	return 0;
}



bool checkMinMax(int16_t *array, uint8_t length, int16_t maxDifference) { // Used to check that the robot is laying still while calibrating
  int16_t min = array[0], max = array[0];
  for (uint8_t i = 1; i < length; i++) {
    if (array[i] < min)
      min = array[i];
    else if (array[i] > max)
      max = array[i];
  }
  return max - min < maxDifference;
}


// Macht nur die Koordinatenzuweisung
// IMU_werteEinlesen muss vorher aufgerufen werden!
// Kippt der Robo nach vorne, liefert die Funktion einen positiven Wert!
float drehgeschwindigkeitUmQuerachse()
{
	return -imu.omegaY;
}
float drehgeschwindigkeitUmHochachse() // rechtsdrehend entsprechend einem nach oben schauenden Vektor
{
	return imu.omegaX;
}



float parameterEinstellen()
{
	if(Serial.available())
	{
		CFG_userInput(Serial.read());
	}
}

// verwendet phiBeschl und phiGyro
// und berechnet phiSelekt
// Dabei wird der Winkel aus dem Beschleunigungssensor stark tiefpassgefiltert, 
// um Schwankungen, hervorgerufen durch Bewegungsbeschleunigung, herauszumitteln. 
// Der integrierte Winkel vom Gyro wird Hochpassgefiltert, um das Driften zu eliminieren. 
static void selektiverFilter(float dt)
{
	float phiB = P.phiBeschl.v;
	float phiG = P.phiGyro.v;

	static float phiHP=0; // Hochpass 
	static float phiTP=0; // Tiefpass
	static float phiG_alt;
	float deltaPhiGyro = phiG - phiG_alt;

	static bool ersterBesuch=true;
	static const float tau = 1; // in s, Filterzeitkonstante

	if(ersterBesuch)
	{
		ersterBesuch = false;
		phiHP = phiB;
		phiTP = phiB;
		deltaPhiGyro = 0;
	}

	phiTP = phiTP + (phiB-phiTP)*dt/tau;
	phiHP = phiHP + deltaPhiGyro - phiHP*dt/tau;

	phiG_alt = phiG;

	P.phiSelekt.v = phiTP + phiHP; 
}

float posSoll(float dt, bool reset)
{
	//return 0;

	const float vMax = P.kMotor.v/7; // maximale sollgeschwindigkeit = 50% der maximal möglichen Geschwindigkeit
		// die Motoren benötigen noch eine Reserve, um ausbalancieren zu können. 

	static float x=0; // soll weg in m (zeitabhängig)

	if(reset)
		x=0;

	static bool vorwaerts=true;
	// if(vorwaerts && x < 1)
	// {
	// 	x += vMax*dt;
	// }
	// if(!vorwaerts && x > 0)
	// {
	// 	x -= vMax*dt;
	// }

	x += vMax*dt;
	return x;


	// // zyklisches Umschalten zwischen vorwärts und rückwärts
	// if(vorwaerts && x >= 1)
	// 	vorwaerts = false;
	// if(!vorwaerts && x <= 0)
	// 	vorwaerts = true;


	//x = x + (vMax*sin(2*M_PI*0.05*dt))*dt;

	// Tiefpassfilter 2. Ordnung, sodass die Trajektorie schön glatt ist:
	// x --> a --> b
	static float a=0;
	static float aAlt=0; 
	static float b=0;
	static float bAlt=0;
	const float tau = 3;

	if(reset)
	{
		a=0; aAlt=0;
		b=0; bAlt=0;
	}

	a = aAlt + (x-aAlt)/tau*dt;
	b = bAlt + (a-bAlt)/tau*dt;

	aAlt = a;
	bAlt = b;

	return b;
}

// Diese Funktion liefert eine soll-Winkelgeschwindigkeit in °/s zurück. 
// positiv = Linkskurve, negativ = Rechtskurve
float lenken()
{
	static bool links = false;
	static bool rechts = false;

	Sonar_measure(sonarL, 150);
	Sonar_measure(sonarR, 150);
	links = sonarL->distance_mm > 0;
	rechts = sonarR->distance_mm > 0;

	if(links && !rechts) // wenn links hinderniss, nach links ausweichen
	{
		return +75;
	}
	if(!links && rechts) // wenn rechts hinderniss, nach rechts ausweichen
	{
		return -75;
	}

	// sonst: richtung beibehalten
	return 0;
}